"""Trigger definitions"""
from tekinstr.instrument import InstrumentSubsystem
from tekinstr.trigger import TriggerBase
from tekinstr.common import validate


class Trigger(TriggerBase, kind="Trigger"):
    """Trigger class

    Attributes:
        owner (Instrument)
        visa (pyvisa.resources.Resource): pyvisa resource
        designation (str): A or B
    """

    def __init__(self, owner, designation):
        super().__init__(owner, designation)
        self.edge = EdgeTrigger(self)
        if designation == "A":
            self.logic = LogicTrigger(self)

    def __dir__(self):
        attrs = super().__dir__()
        if self.kind == "EDGE":
            attrs.remove("logic")
        else:
            attrs.remove("edge")
        return attrs


class EdgeTrigger(InstrumentSubsystem, kind="Edge"):
    """Edge trigger class

    Attributes:
        owner (Instrument)
    """

    def __init__(self, owner):
        super().__init__(owner)
        self._designation = owner._designation

    @property
    def coupling(self):
        """value (str): {AC, DC, HFRej, LFRej, NOISErej}"""
        return self._visa.query(f"TRIGGER:{self._designation}:EDGE:COUPLING?")

    @coupling.setter
    @validate
    def coupling(self, value):
        self._visa.write(f"TRIGGER:{self._designation}:EDGE:COUPLING {value}")

    @property
    def slope(self):
        """value (str): trigger on slope, {RISE, FALL, EITHER}"""
        return self._visa.query(f"TRIGGER:{self._designation}:EDGE:SLOPE?")

    @slope.setter
    @validate
    def slope(self, value):
        self._visa.write(f"TRIGGER:{self._designation}:EDGE:SLOPE {value}")

    @property
    def source(self):
        """value (str): trigger source, {AUX, CHx, Dx, LINE, RF}"""
        return self._visa.query(f"TRIGGER:{self._designation}:EDGE:SOURCE?")

    @source.setter
    @validate
    def source(self, value):
        self._visa.write(f"TRIGGER:{self._designation}:EDGE:SOURCE {value}")


class LogicTrigger(InstrumentSubsystem, kind="Logic"):
    """Logic trigger class

    Attributes:
        owner (Instrument)
    """

    def __init__(self, owner):
        super().__init__(owner)
        self._designation = "A"
        n_channels = self._instr.features["ANALOG:NUMCHANNELS"]
        self._channels = [LogicInputChannel(self, n) for n in range(1, n_channels + 1)]
        for x, ch in enumerate(self._channels, 1):
            setattr(self, f"ch{x}", ch)
        self.clock = LogicClock(self)
        self.pattern = LogicPattern(self)

    @property
    def function(self):
        """value (str): {AND, NAND, NOR, OR}"""
        q_str = f"TRIGGER:A:LOGIC:FUNCTION?"
        return self._visa.query(q_str)

    @function.setter
    @validate
    def function(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:FUNCTION {value}"
        self._visa.write(cmd_str)

    def __dir__(self):
        attrs = super().__dir__()
        if self.clock.source is not None:
            attrs.remove("pattern")
        return attrs


class LogicInputChannel(InstrumentSubsystem, kind="LogicInput"):
    """Logic input channel state and threshold

    Attributes:
        owner (Instrument)
        visa (pyvisa.resources.Resource): pyvisa resource
        CHx (int): input channel
    """

    def __init__(self, owner, CHx):
        super().__init__(owner)
        self._designation = "A"
        self._ch = CHx

    def __repr__(self):
        return f"<{self._designation} trigger: {self._kind}>"

    @property
    def state(self):
        """value (str): input condition, {HIGH, LOW, X}"""
        q_str = f"TRIGGER:A:LOGIC:INPUT:CH{self._ch}?"
        return self._visa.query(q_str)

    @state.setter
    @validate
    def state(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:INPUT:CH{self._ch} {value}"
        self._visa.write(cmd_str)

    @property
    def threshold(self):
        """value (float or str): {voltage, ECL, TTL}"""
        q_str = f"TRIGGER:A:LOGIC:THRESHOLD:CH{self._ch}?"
        value = self._visa.query(q_str)
        try:
            value = float(value)
        except ValueError:
            pass
        return value

    @threshold.setter
    @validate
    def threshold(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:THRESHOLD:CH{self._ch} {value}"
        self._visa.write(cmd_str)


class LogicClock(InstrumentSubsystem, kind="LogicClock"):
    """Logic clock source and edge

    Attributes:
        owner (Instrument)
    """

    def __init__(self, owner):
        super().__init__(owner)
        self._designation = "A"

    def __repr__(self):
        return f"<{self._designation} trigger: {self._kind}>"

    @property
    def source(self):
        """value (str): clock source, {CHx, Dx, RF, NONE}"""
        q_str = "TRIGGER:A:LOGIC:INPUT:CLOCK:SOURCE?"
        value = self._visa.query(q_str)
        return None if value == "NONE" else value

    @source.setter
    @validate
    def source(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:INPUT:CLOCK:SOURCE {value}"
        self._visa.write(cmd_str)

    @property
    def edge(self):
        """value (str): {RISE, FALL}"""
        q_str = "TRIGGER:A:LOGIC:INPUT:CLOCK:EDGE?"
        return self._visa.query(q_str)

    @edge.setter
    @validate
    def edge(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:INPUT:CLOCK:EDGE {value}"
        self._visa.write(cmd_str)


class LogicPattern(InstrumentSubsystem, kind="Logic Pattern"):
    """Logic pattern

    Attributes:
        owner (Instrument)
    """

    @property
    def when(self):
        """value (str): {TRUE, FALSE, LESSTHAN, MORETHAN, EQUAL, UNEQUAL}"""
        q_str = "TRIGGER:A:LOGIC:PATTERN:WHEN?"
        return self._visa.query(q_str)

    @when.setter
    @validate
    def when(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:PATTERN:WHEN {value}"
        self._visa.write(cmd_str)

    @property
    def deltatime(self):
        """value (float): trigger delta time in seconds"""
        q_str = "TRIGGER:A:LOGIC:PATTERN:DELTATIME?"
        return float(self._visa.query(q_str))

    @deltatime.setter
    @validate
    def deltatime(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:PATTERN:DELTATIME {value}"
        self._visa.write(cmd_str)


class SetHoldTrigger(InstrumentSubsystem, kind="SetHold"):
    """Logic setup & hold trigger class

    Attributes:
        owner (Instrument)
    """

    def __init__(self, owner):
        super().__init__(owner)
        self._designation = "A"

    def __repr__(self):
        return f"<{self._designation} trigger: {self._kind}>"



"""Oscilloscope base class"""
import re
from datetime import datetime, timedelta
import sys
import asyncio
import itertools
from dataclasses import dataclass
import numpy as np
import pyvisa
from tekinstr.instrument import Instrument, InstrumentSubsystem
from tekinstr.common import validate
from waveformDT.waveform import WaveformDT

# pylint: disable=invalid-name
@dataclass
class State:
    """Signal for controlling asyncio tasks"""

    running: bool = False


class OscilloscopeBase(Instrument, kind="OscilloscopeBase"):
    """Oscilloscope base class

    Attributes:
        instr (Model)
    """

    def __init__(self, instr):
        super().__init__(instr)
        self._state = State()

    @property
    def horizontal_scale(self):
        """value (float): amount of time per division coerced to nearest setting"""
        return float(self._visa.query("HORIZONTAL:SCALE?"))

    @horizontal_scale.setter
    @validate
    def horizontal_scale(self, value):
        self._visa.write(f"HORIZONTAL:SCALE {float(value)}")

    @property
    def horizontal_delay_mode(self):
        """value (str): delay_time or pretrigger_percent
        delay_time: center waveform horizontal_position seconds after trigger occurs
        pretrigger_percent: horizontal_position percent of record length
        """
        mode = self._visa.query("HORIZONTAL:DELAY:STATE?")
        return "delay_time" if mode in ["ON", "1"] else "pretrigger_percent"

    @horizontal_delay_mode.setter
    @validate
    def horizontal_delay_mode(self, value):
        mode = "ON" if value == "delay_time" else "OFF"
        self._visa.write(f"HORIZONTAL:DELAY:STATE {mode}")

    @property
    def horizontal_position(self):
        """value (float): position of waveform on display based on horizontal_delay_mode
        delay_time: center waveform horizontal_position seconds after trigger occurs
        pretrigger_percent: horizontal_position percent of record length
        """
        delay_time = float(self._visa.query("HORIZONTAL:DELAY:TIME?"))
        pretrigger_percent = float(self._visa.query("HORIZONTAL:TRIGGER:POSITION?"))
        if self.horizontal_delay_mode in ["0", "OFF"]:
            horizontal_position = pretrigger_percent
        else:
            horizontal_position = delay_time
        return horizontal_position

    @horizontal_position.setter
    @validate
    def horizontal_position(self, value):
        if self.horizontal_delay_mode in ["0", "OFF"]:
            self._visa.write(f"HORIZONTAL:TRIGGER:POSITION {float(value)}")
        else:
            self._visa.write(f"HORIZONTAL:DELAY:TIME {float(value)}")

    @property
    def record_length(self):
        """value (int): 500 or 10000 samples"""
        return int(self._visa.query("HORIZONTAL:RECORDLENGTH?"))

    @record_length.setter
    @validate
    def record_length(self, value):
        self._visa.write(f"HORIZONTAL:RECORDLENGTH {int(value)}")

    @property
    def acquisition_mode(self):
        """value (str): acquisition mode; SAMPLE, PEAKDETECT, AVERAGE, ENVELOPE
        AVERAGE - specify num_averages
        ENVELOPE - specify num_envelopes
        """
        return self._visa.query("ACQUIRE:MODE?")

    @acquisition_mode.setter
    @validate
    def acquisition_mode(self, value):
        self._visa.write(f"ACQUIRE:MODE {value}")

    @property
    def acquisition_count(self):
        """(int): number of acquisitions since setting acquisition_state to RUN

        This value is reset to zero when any acquisition, horizontal or vertical
        arguments that affect the waveform are changed.
        """
        return int(self._visa.query("ACQUIRE:NUMACQ?"))

    @property
    def num_averages(self):
        """value (int): number of waveforms to average
        The range of values is 2 to 512 in powers of 2.
        """
        return int(self._visa.query("ACQUIRE:NUMAVG?"))

    @num_averages.setter
    @validate
    def num_averages(self, value):
        self._visa.write(f"ACQUIRE:NUMAVG {value}")

    @property
    def num_envelopes(self):
        """value (int or str): number of envelopes from 1 to 2000 or INFINITE"""
        return self._visa.query("ACQUIRE:NUMENV?")

    @num_envelopes.setter
    @validate
    def num_envelopes(self, value):
        self._visa.write(f"ACQUIRE:NUMENV {value}")

    @property
    def acquisition_state(self):
        """value (str): start or stop acquisition; RUN (1) or STOP (0)"""
        return self._visa.query("ACQUIRE:STATE?")

    @acquisition_state.setter
    @validate
    def acquisition_state(self, value):
        self._visa.write(f"ACQUIRE:STATE {value}")

    @property
    def single_acquisition(self):
        """value (bool): single or continuous acquisitions
        For single acqusitions set to True
        For continuous acquisitions set to False
        """
        stop_after = self._visa.query("ACQUIRE:STOPAFTER?")
        return stop_after == "SEQUENCE"

    @single_acquisition.setter
    @validate
    def single_acquisition(self, value):
        stop_after = "SEQUENCE" if value else "RUNSTOP"
        self._visa.write(f"ACQUIRE:STOPAFTER {stop_after}")
        stop_after = self._visa.query("ACQUIRE:STOPAFTER?")

    def _get_wfmpre(self, inout=""):
        """Get waveform preamble

        Args:
            inout (str): IN, OUT or '' depending on model and signal direction
        """
        self._visa.write("HEADER ON")
        pre_str = self._visa.query(f"WFM{inout}PRE?").replace(f":WFM{inout}PRE:", "")
        raw_preamble = dict([kv.split(" ", maxsplit=1) for kv in pre_str.split(";")])
        preamble = {}
        for k, v in raw_preamble.items():
            value = v.strip('"')
            try:
                value = int(v)
            except ValueError:
                try:
                    value = float(v)
                except ValueError:
                    pass
            preamble[k] = value
        self._visa.write("HEADER OFF")
        return preamble

    async def _show_spinner(self):
        """Show an in-progress spinner during acquisition"""
        spinner = itertools.cycle(["-", "/", "|", "\\"])
        try:
            while self._state.running:
                sys.stdout.write(next(spinner))
                sys.stdout.flush()
                sys.stdout.write("\b")
                await asyncio.sleep(0.5)
            return 0
        except asyncio.CancelledError:
            pass
        finally:
            sys.stdout.write("\b \b")

    async def _acquire(self):
        """Acquire single sequence"""
        try:
            self._visa.write("*SRE 32")
            self._visa.write("*ESE 61")
            self._visa.write("*CLS")
            original_sa = self.single_acquisition
            self.single_acquisition = True
            self._visa.write("ACQUIRE:STATE RUN; *OPC")
            # poll the ESB bit for an event occurance indicating completion or error
            while not self._visa.stb & 32:
                await asyncio.sleep(1)
            esr = int(self._visa.query("*ESR?"))
            op_complete = bool(esr & 1)
            self.single_acquisition = original_sa
            return 0 if op_complete else 1
        except asyncio.CancelledError:
            pass
        except pyvisa.VisaIOError as exc:
            if exc.abbreviation == "VI_ERROR_TMO":
                raise TimeoutError(
                    "Acquisition timed out due to loss of communication"
                ) from None
            else:
                raise
        finally:
            self._state.running = False

    async def _start_task(self, timeout):
        """timeout (int): timeout in seconds"""
        self._state.running = True
        task = asyncio.gather(self._show_spinner(), self._acquire())
        try:
            ret_value = await asyncio.wait_for(task, timeout)
        except asyncio.TimeoutError:
            task.exception()  # retrieve the _GatheringFuture exception
            raise TimeoutError(
                "Acquisition didn't complete before specified timeout value"
            ) from None
        else:
            return ret_value

    def read(self, channels, samples="all", timeout=10, wdt=True, previous=True):
        """Transfer waveform data of the specified channel(s) from the oscilloscope

        Args:
            channels (str): specification such as 'CH1', 'CH1:4', or ['CH1', 'CH2']
                Acquisition is temporarily stopped to ensure all channel data come from
                the same acquisition instance.
            samples ('all', int or tuple): If 'all', return record_length samples.
                If an integer, return N samples starting with the
                first sample in the record. If a tuple (start, stop), return
                samples start to stop from the record.
            timeout (float): timeout, in seconds, only has meaning when previous
                is set to False, otherwise return the latest acquired waveform. A value
                of None means no timeout.
            wdt (bool): If true, return data as WaveformDT.
            previous (bool): If True, just read the existing waveform data. If False,
                initiate a new single sequence.
        Returns:
            WaveformDT or numpy.ndarray
        """
        chs = []
        match = re.match("^CH(?P<first>[1-4])(?P<last>:[1-4])?|(MATH)$", channels)
        if match is None:
            raise ValueError(f"{channels} not a valid specification")
        if match.groups()[2] is not None:
            channels = [channels]
        else:
            first = int(match.group("first"))
            last = match.group("last")
            if last is None:
                last = first
            else:
                last = int(last[-1])
            for ch in range(first, last + 1):
                chs.append(f"CH{ch}")
            channels = chs
        if isinstance(samples, tuple):
            start, stop = samples
        elif isinstance(samples, int):
            start = 1
            stop = samples
        else:
            start = 1
            stop = self.record_length
        self._visa.write(f"DATA:START {start}")
        self._visa.write(f"DATA:STOP {stop}")
        self._visa.write("DATA:WIDTH 1")
        self._visa.write("DATA:ENCDG RIBinary")
        original_state = self.acquisition_state
        self._visa.write("ACQUIRE:STATE STOP")
        if not previous:
            ret_value = asyncio.run(self._start_task(timeout))
            if ret_value is None:
                return None
            if ret_value[1] > 0:
                print(self._instr.status.event_status)
                return None
        time = self._visa.query("TIME?")
        date = self._visa.query("DATE?")
        data = []
        for channel in channels:
            self._visa.write(f"DATA:SOURCE {channel}")
            preamble = self._get_wfmpre()
            y_offset = preamble["YZERO"]  # in y_unit
            y_multiplier = preamble[
                "YMULT"
            ]  # conversion from data point level to y_unit
            y_position = preamble["YOFF"]  # in data point levels
            rawdata = self._visa.query_binary_values(
                "CURVE?", "b", container=np.ndarray
            )
            data.append(y_offset + y_multiplier * (rawdata - y_position))
        self.acquisition_state = original_state
        data = data[0] if len(data) == 1 else np.asarray(data)
        if wdt:
            time = [int(n) for n in time.strip().strip('"').split(":")]
            date = [int(n) for n in date.strip().strip('"').split("-")]
            t0 = datetime(*date, *time) + timedelta(seconds=preamble["XZERO"])
            dt = preamble["XINCR"]
            data = WaveformDT(data, dt, t0)
            setattr(data, "wf_start_offset", preamble["XZERO"])
            setattr(data, "_xunit", preamble["XUNIT"])
            setattr(data, "_yunit", preamble["YUNIT"])
            setattr(data, "y_position", y_position * y_multiplier)
            setattr(data, "y_offset", y_offset)
        return data

    def force_trigger(self):
        """Force a trigger event"""
        self._visa.write("TRIGGER FORCE")

    @property
    def trigger_state(self):
        """(str): ARMED, AUTO, READY, SAVE, TRIGGER"""
        return self._visa.query("TRIGGER:STATE?")


class ChannelBase(InstrumentSubsystem, kind="ChannelBase"):
    """Oscilloscope channel

    Attributes:
        owner (Instrument)
        CHx (int): channel number
    """

    def __init__(self, owner, CHx):
        super().__init__(owner)
        self._ch = CHx

    @property
    def bandwidth(self):
        """value (str): TWENTY, ONEFIFTY or FULL"""
        return self._visa.query(f"CH{self._ch}:BANDWIDTH?")

    @bandwidth.setter
    def bandwidth(self, value):
        self._visa.write(f"CH{self._ch}:BANDWIDTH {value}")

    @property
    def coupling(self):
        """value (str): AC, DC or GND"""
        return self._visa.query(f"CH{self._ch}:COUPLING?")

    @coupling.setter
    def coupling(self, value):
        self._visa.write(f"CH{self._ch}:COUPLING {value}")

    @property
    def invert(self):
        """value (str): ON (1) or OFF (0)"""
        return self._visa.query(f"CH{self._ch}:INVERT?")

    @invert.setter
    def invert(self, value):
        self._visa.write(f"CH{self._ch}:INVERT {value}")

    @property
    def offset(self):
        """value (float): value subtracted from the signal before acquisition"""
        return float(self._visa.query(f"CH{self._ch}:OFFSET?"))

    @offset.setter
    def offset(self, value):
        self._visa.write(f"CH{self._ch}:OFFSET {value}")

    @property
    def position(self):
        """value (float): vertical position in divisions from center graticule"""
        return float(self._visa.query(f"CH{self._ch}:POSITION?"))

    @position.setter
    def position(self, value):
        self._visa.write(f"CH{self._ch}:POSITION {value}")

    @property
    def scale(self):
        """value (float): vertical gain in y_unit per division"""
        return float(self._visa.query(f"CH{self._ch}:SCALE?"))

    @scale.setter
    def scale(self, value):
        self._visa.write(f"CH{self._ch}:SCALE {value}")

    @property
    def y_unit(self):
        """value (str): 'V' for voltage or 'A' for amperage"""
        return self._visa.query(f"CH{self._ch}:YUN?").strip('"')

    @y_unit.setter
    def y_unit(self, value):
        self._visa.write(f"CH{self._ch}:YUN '{value}'")


class ProbeBase(InstrumentSubsystem, kind="ProbeBase"):
    """Oscilloscope channel

    Attributes:
        owner (InstrumentSubsystem)
    """

    def __init__(self, owner):
        super().__init__(owner)
        self._ch = owner._ch

    @property
    def model(self):
        """(str): probe model"""
        return self._visa.query(f"CH{self._ch}:ID?").strip('"')

    @property
    def gain(self):
        """value (float): probe gain factor (output/input)"""
        return float(self._visa.query(f"CH{self._ch}:PROBE?"))

    @gain.setter
    @validate
    def gain(self, value):
        self._visa.write(f"CH{self._ch}:PROBE {value}")

    @property
    def impedance(self):
        """value (float): probe impedance in ohms"""
        impedance = self._visa.query(f"CH{self._ch}:IMPEDANCE?")
        if impedance == "MEG":
            impedance = 1e6
        elif impedance == "FIFTY":
            impedance = 50.0
        return impedance

    def __repr__(self):
        instr_model = self._instr.model
        instr = self._owner._owner._kind
        subsys = self._owner._kind
        return f"<{instr_model} {instr} {subsys} {self._kind} {self.model}>"


class MathChannelBase(InstrumentSubsystem, kind="MathChannelBase"):
    """Base math channel

    Parameters
    ----------
        owner :Instrument
    """

    @property
    def definition(self):
        """The math definition

        Parameters
        ----------
        value : str
            formula such as 'CH2-MEAN(CH2)'
        """
        return self._visa.query("MATH?")

    @definition.setter
    @validate
    def definition(self, value):
        self._visa.write(f'MATH:DEFINE "{value}"')



"""Test MSO4041B"""
# pylint: disable=all
import matplotlib.pyplot as plt
from tekinstr import CommChannel

cc = CommChannel("10.2.128.17")
tek = cc.get_instrument()
data = tek.oscilloscope.read("CH1")
plt.plot(*data.to_xy())
plt.show()
cc.close()



"""Base file system definition"""
from tekinstr.instrument import Instrument
from tekinstr.common import validate


class FileSystemBase(Instrument, kind="FileSystemBase"):
    """File system

    Attributes:
        instr (Model)
    """

    @property
    def listing(self):
        """Return the current directory listing"""
        result = self._visa.query("FILESYSTEM?")
        raw_listing, _ = result.split(";")
        listing = raw_listing.replace('"', "").split(",")
        return listing

    @property
    def cwd(self):
        """Current working directory

        Parameters
        ----------
        value : str
            change working directory to 'value' e.g. 'E:/data'
        """
        result = self._visa.query("FILESYSTEM:CWD?")
        return result.replace('"', "")

    @cwd.setter
    @validate
    def cwd(self, value):
        self._visa.write(f"FILESYSTEM:CWD '{value}'")

    def mkdir(self, directory):
        """Make 'directory' at current working directory

        Parameters
        ----------
        directory : str
        """
        self._visa.write(f"FILESYSTEM:MKDIR '{directory}'")



"""Test WaveformDT"""
from datetime import datetime
import numpy as np
from pytest import approx
from unyt import s, m, unyt_array  # pylint: disable=no-name-in-module
from .waveform import WaveformDT

# pylint: disable=missing-docstring
# pylint: disable=pointless-statement
# pylint: disable=no-member
# pylint: disable=invalid-name
def test_attributes_nounits():
    wf = WaveformDT([0, 1, 2], 1, 0)
    assert isinstance(wf.Y, np.ndarray)
    assert wf.dt == 1
    assert wf.t0 == 0
    assert wf.yunit == None
    assert wf.xunit == None
    assert wf.size == 3
    x, y = wf.to_xy()
    assert (x == np.asarray([0.0, 1.0, 2.0])).all()
    assert (y == np.asarray([0, 1, 2])).all()
    wf = WaveformDT(list(range(10)), 1, 0)
    assert (wf.head().Y == np.arange(5)).all()
    assert (np.asarray(wf.tail()) == np.arange(5, 10)).all()
    x, _ = wf.tail().to_xy()
    assert (x == np.arange(5.0, 10.0)).all()
    assert wf.min == 0
    assert wf.max == 9
    assert wf.std == approx(2.8722813232690143)
    wf.set_attributes(attr="custom")
    assert wf.attr == "custom"
    assert (wf["x"] == np.arange(10.0)).all()
    assert (wf["y"] == np.arange(10)).all()


def test_math_nounits():
    wf = WaveformDT([0, 1, 2], 1, 0)
    assert (np.asarray(2 * wf) == np.asarray([0, 2, 4])).all()
    assert (np.asarray(wf + 2) == np.asarray([2, 3, 4])).all()
    np.multiply.at(wf, [0, 1, 2], 2.0)
    assert (wf.Y == np.asarray([0.0, 2.0, 4.0])).all()
    arr_out = np.asarray([0.0] * 3)
    np.multiply(wf, 2.0, out=arr_out)
    assert (arr_out == np.asarray([0.0, 4.0, 8.0])).all()
    wf_out = WaveformDT([0.0] * 3, 1, 0)
    assert isinstance(wf_out, WaveformDT)
    np.multiply(wf, 2.0, out=wf_out)
    assert (wf_out.Y == np.asarray([0.0, 4.0, 8.0])).all()


def test_attributes_withunits():
    wf = WaveformDT([0, 1, 2], 1, 0)
    wf.yunit = "m"
    wf.xunit = "s"
    x, y = wf.to_xy()
    assert (x == unyt_array([0.0, 1.0, 2.0], "s")).all()
    assert (y == unyt_array([0, 1, 2], "m")).all()
    wf = WaveformDT([0, 1, 2] * m, 1 * s, 0 * s)
    assert wf.yunit == m
    assert wf.xunit == s
    x, y = wf.to_xy(inunits=False)
    assert (x == np.arange(3.0)).all()
    assert (y == np.arange(3)).all()


def test_absolute_xy():
    t = lambda s: datetime(2020, 1, 1, 0, 0, s)
    wf = WaveformDT([0, 1, 2], 1, t(0))
    x, _ = wf.to_xy(relative=False)
    expected = np.asarray([t(s) for s in range(3)], dtype="datetime64[ns]")
    assert (x == expected).all()



"""Test Model"""
import os
import pytest
from tekinstr import CommChannel

# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
@pytest.fixture(scope="module")
def tek(request):
    path = os.path.dirname(os.path.abspath(__file__))
    os.environ["PYVISA_LIBRARY"] = path + r"\sim_hw.yml@sim"
    address = request.config.getoption("ipaddress")
    cc = CommChannel(address)
    yield cc.get_instrument()
    cc.close()


def test_model(tek):
    assert tek.model == "MDO3024"



"""Test MDO3024"""
import pytest
from tekinstr import CommChannel

# pylint: disable=missing-function-docstring
# pylint: disable=redefined-outer-name
@pytest.fixture(scope="module")
def tek(request):
    addr = request.config.getoption("ipaddress")
    cc = CommChannel(addr)
    yield cc.get_instrument()
    cc.close()


def test(tek):
    assert tek.model == "MDO3024"



"""Oscilloscope instrument within TDS3000 Series"""
from tekinstr.oscilloscope import OscilloscopeBase, ChannelBase, ProbeBase
from tekinstr.tds3000.trigger import Trigger
from tekinstr.measurement import Measurement


class Oscilloscope(OscilloscopeBase, kind="Oscilloscope"):
    """Oscilloscope instrument within TDS3000 Series"""

    def __init__(self, instr, n_channels):
        super().__init__(instr)
        self._channels = [Channel(self, x) for x in range(1, n_channels + 1)]
        for x, ch in enumerate(self._channels, 1):
            setattr(self, f"ch{x}", ch)
        self.trigger = Trigger(self, "A")
        self.B_trigger = Trigger(self, "B")
        self.measurement = Measurement(self, 4)


class Channel(ChannelBase, kind="CH"):
    """Channel"""

    def __init__(self, owner, CHx):
        super().__init__(owner, CHx)
        self._probe = Probe(self)

    @property
    def probe(self):
        """(Probe): probe"""
        return self._probe

    def __repr__(self):
        owner = self._owner.__repr__().strip("<>")
        return f"<{owner} {self._kind}{self._ch}>"


class Probe(ProbeBase, kind="Probe"):
    """Probe"""

    @property
    def gain(self):
        """value (float): probe attenuation factor (input/output)"""
        return float(self._visa.query(f"CH{self._ch}:PROBE?"))

    def __repr__(self):
        owner = self._owner.__repr__().strip("<>")
        return f"<{owner} {self._kind} {self.model}>"



"""Trigger base definitions"""
from tekinstr.instrument import InstrumentSubsystem
from tekinstr.common import validate


class TriggerBase(InstrumentSubsystem, kind="TriggerBase"):
    """Trigger base class for all trigger types

    Attributes:
        owner (Instrument)
        designation (str): A or B
    """

    def __init__(self, owner, designation):
        super().__init__(owner)
        self._designation = designation

    def __repr__(self):
        return f"<{self._instr.model} {self._owner._kind} {self._designation}-{self._kind}>"

    @property
    def kind(self):
        """value (str): type of trigger; {EDGE, LOGIC, PULSE}"""
        return self._visa.query(f"TRIGGER:{self._designation}:TYPE?")

    @kind.setter
    @validate
    def kind(self, value):
        if self._designation == "B":
            raise AttributeError("can't set B trigger type")
        self._visa.write(f"TRIGGER:{self._designation}:TYPE {value}")

    @property
    def parameters(self):
        """(dict): trigger parameters"""
        return self._get_parameters()

    def _get_parameters(self):
        self._visa.write("HEADER ON")
        raw_str = self._visa.query(f"TRIGGER:{self._designation}?")
        self._visa.write("HEADER OFF")
        parameters = dict([p.split(" ", maxsplit=1) for p in raw_str.split(";")])
        for k, v in parameters.items():
            try:
                v = int(v)
            except ValueError:
                try:
                    v = float(v)
                except ValueError:
                    pass
            parameters[k] = v
        return parameters

    @property
    def mode(self):
        """value (str): A trigger mode, AUTO or NORMAL"""
        return self._visa.query("TRIGGER:A:MODE?")

    @mode.setter
    @validate
    def mode(self, value):
        self._visa.write(f"TRIGGER:A:MODE {value}")

    @property
    def state(self):
        """value (str): B trigger state, ON or OFF"""
        return self._visa.query("TRIGGER:B:STATE?")

    @state.setter
    @validate
    def state(self, value):
        self._visa.write(f"TRIGGER:B:STATE {value}")

    @property
    def level(self):
        """value (float or str): trigger threshold level, {voltage, ECL, TTL}
            float: level in volts
            str: ECL (-1.3 V) or TTL (1.4 V)
        """
        level = self._visa.query(f"TRIGGER:{self._designation}:LEVEL?")
        try:
            level = float(level)
        except ValueError:
            pass
        return level

    @level.setter
    @validate
    def level(self, value):
        self._visa.write(f"TRIGGER:{self._designation}:LEVEL {value}")

    def set_level(self):
        """set trigger level to 50 %"""
        self._visa.write(f"TRIGGER:{self._designation}:SETLEVEL")

    @property
    def by(self):
        """value (str): delay B trigger by TIME or EVENTS"""
        return self._visa.query("TRIGGER:B:BY?")

    @by.setter
    @validate
    def by(self, value):
        self._visa.write(f"TRIGGER:B:BY {value}")

    @property
    def time(self):
        """value (float): delay B trigger by value seconds when 'by' set to TIME"""
        return float(self._visa.query("TRIGGER:B:TIME?"))

    @time.setter
    @validate
    def time(self, value):
        self._visa.write(f"TRIGGER:B:TIME {value}")

    @property
    def events(self):
        """value (int): B trigger occurs on the value occurance after A trigger"""
        return int(self._visa.query("TRIGGER:B:EVENTS?"))

    @events.setter
    @validate
    def events(self, value):
        self._visa.write(f"TRIGGER:B:EVENTS:COUNT {value}")

    @property
    def holdoff(self):
        """value (float): A trigger holdoff time in seconds"""
        return float(self._visa.query("TRIGGER:A:HOLDOFF?"))

    @holdoff.setter
    @validate
    def holdoff(self, value):
        self._visa.write(f"TRIGGER:A:HOLDOFF:TIME {value}")

    def __dir__(self):
        attrs = super().__dir__()
        if self._designation == "A":
            for attr in ["state", "by", "time", "events"]:
                attrs.remove(attr)
        else:
            for attr in ["mode", "holdoff"]:
                attrs.remove(attr)
        return sorted(attrs)


class LogicInput(InstrumentSubsystem, kind="LogicTrigger:Input"):
    """Logic Input class

    Attributes:
        owner (LogicTrigger)
        input_n (int): logic trigger input, 1 or 2
    """

    def __init__(self, owner, input_n):
        super().__init__(owner)
        self._input_n = input_n

    def __repr__(self):
        instr_model = self._instr.model
        instr = self._owner._owner._kind
        subsys = self._owner._kind
        return f"<{instr_model} {instr} {subsys} {self._kind}>"

    @property
    def level(self):
        """value (str): HIGH or LOW"""
        q_str = f"TRIGGER:A:LOGIC:INPUT{self._input_n}:LOGICLEVEL?"
        return self._visa.query(q_str)

    @level.setter
    @validate
    def level(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:INPUT{self._input_n}:LOGICLEVEL {value}"
        self._visa.write(cmd_str)

    @property
    def slope(self):
        """value (str): input 2 state logic trigger slope, RISE or FALL"""
        q_str = f"TRIGGER:A:LOGIC:INPUT2:SLOPE?"
        return self._visa.query(q_str)

    @slope.setter
    @validate
    def slope(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:INPUT2:SLOPE {value}"
        self._visa.write(cmd_str)

    def __dir__(self):
        attrs = super().__dir__()
        if self._input_n == 1:
            attrs.remove("slope")
        return attrs

    @property
    def source(self):
        """value (str): input signal source, CHx"""
        q_str = f"TRIGGER:A:LOGIC:INPUT{self._input_n}:SOURCE?"
        return self._visa.query(q_str)

    @source.setter
    @validate
    def source(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:INPUT{self._input_n}:SOURCE {value}"
        self._visa.write(cmd_str)

    @property
    def threshold(self):
        """value (float or str): threshold in volts, ECL or TTL"""
        q_str = f"TRIGGER:A:LOGIC:INPUT{self._input_n}:THRESHOLD?"
        threshold_str = self._visa.query(q_str)
        try:
            threshold = float(threshold_str)
        except ValueError:
            threshold = threshold_str
        return threshold

    @threshold.setter
    @validate
    def threshold(self, value):
        cmd_str = f"TRIGGER:A:LOGIC:INPUT{self._input_n}:THRESHOLD {value}"
        self._visa.write(cmd_str)



"""Spectrum analyzer instrument definition"""
from tekinstr.spectrum_analyzer import SpectrumAnalyzerBase


class SpectrumAnalyzer(SpectrumAnalyzerBase, kind="Spectrum Analyzer"):
    """Spectrum analyzer instrument

    Attributes:
        instr (Model)
        visa (pyvisa.resources.Resource): pyvisa resource
    """



"""File system"""
import pathlib
from tekinstr.filesystem import FileSystemBase


class FileSystem(FileSystemBase, kind="FileSystem"):
    """File system

    Attributes:
        instr (Model)
    """

    def mount(self, path, drive, username="", password=""):
        """Mount a network drive at 'path' to instrument file system 'drive'

        Only supports Microsoft Windows NFS or CIFS networks.
        Caution: username and password are in-the-clear

        Parameters
        ----------
        path : str
            network path e.g. '<server>/this/that/mydir'
        drive : str
            instrument file system, network drives start at 'G:'
        username : str
        password : str
        """
        parts = pathlib.PurePath(path).parts
        server, path = (parts[0], "/".join(parts[1:]))
        cmd = ";".join([drive, server, path, username, password])
        self._visa.write(f"FILESYSTEM:MOUNT:DRIVE '{cmd}'")

    def unmount(self, drive):
        """Unmount network 'drive'

        Parameters
        ----------
        drive : str
            drive to unmount e.g. 'G:'
        """
        self._visa.write(f"FILESYSTEM:UNMOUNT:DRIVE '{drive}'")
